# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from typing import Any, BinaryIO, Dict, Tuple

from austin.format.pprof.profile_pb2 import Profile, Sample
from austin.stats import Frame, Sample as AustinSample


PROCESS_ID_LABEL = "Process ID"
THREAD_ID_LABEL = "Thread ID"

TIME_TYPE = "Sampling time"
TIME_UNIT = "μs"
MEMORY_TYPE = "Memory"
MEMORY_ALLOC_TYPE = "Allocated memory"
MEMORY_DEALLOC_TYPE = "Deallocated memory"
MEMORY_UNIT = "B"


class Mode(Enum):
    """Austin profiling mode."""

    TIME = 0
    MEMORY = 1
    FULL = 2


class PProf:
    """PProf generator class.

    Used to convert Austin collapsed format into the pprof protobuf format.
    """

    def __init__(self, mode: Mode = Mode.TIME) -> None:
        self._string_table: Dict[str, int] = {}
        self._location_map: Dict[Frame, int] = {}
        self._function_map: [Dict[Tuple[str, str]], int] = {}

        # Create the protobuf Profile message
        self.profile = Profile()
        self.profile.string_table.append("")

        # Add sample types according to the mode
        if mode == Mode.MEMORY:
            _ = self.profile.sample_type.add()
            _.type = self.get_string(MEMORY_TYPE)
            _.unit = self.get_string(MEMORY_UNIT)

        elif mode in (Mode.TIME, Mode.FULL):
            _ = self.profile.sample_type.add()
            _.type = self.get_string(TIME_TYPE)
            _.unit = self.get_string(TIME_UNIT)

        if mode == Mode.FULL:
            _ = self.profile.sample_type.add()
            _.type = self.get_string(MEMORY_ALLOC_TYPE)
            _.unit = self.get_string(MEMORY_UNIT)

            _ = self.profile.sample_type.add()
            _.type = self.get_string(MEMORY_DEALLOC_TYPE)
            _.unit = self.get_string(MEMORY_UNIT)

        self.mode = mode

    def get_string(self, string: str) -> int:
        """Get the string table index for the given string."""
        try:
            return self._string_table[string]
        except KeyError:
            index = len(self.profile.string_table)
            self._string_table[string] = index
            self.profile.string_table.append(string)
            return index

    def add_label_to_sample(self, sample: Sample, key: Any, value: Any) -> None:
        """Add a sample label to the given sample.

        The ``key`` and ``value`` arguments are both converted to strings.
        """
        _ = sample.label.add()
        _.key = self.get_string(str(key))
        _.str = self.get_string(str(value))

    def get_function(self, frame: Frame) -> int:
        """Get the function id from the given Austin frame."""
        key = (frame.function, frame.filename)
        try:
            return self._function_map[key]
        except KeyError:
            function = self.profile.function.add()
            function.id = len(self.profile.function)
            function.name = self.get_string(frame.function)
            function.filename = self.get_string(frame.filename)

            self._function_map[key] = function.id

            return function.id

    def get_location(self, frame: Frame) -> int:
        """Get the location id from the given Austin frame."""
        try:
            return self._location_map[frame]
        except KeyError:
            location = self.profile.location.add()
            location.id = len(self.profile.location)
            line = location.line.add()
            line.function_id = self.get_function(frame)
            line.line = frame.line

            self._location_map[frame] = location.id

            return location.id

    def add_sample(self, sample: AustinSample) -> None:
        """Add a sample to the pprof generator."""
        # Create new pprof sample
        pprof_sample = self.profile.sample.add()

        # Add process and thread id labels
        self.add_label_to_sample(pprof_sample, THREAD_ID_LABEL, sample.thread)
        self.add_label_to_sample(pprof_sample, PROCESS_ID_LABEL, sample.pid)

        # Add metrics
        if self.mode in (Mode.TIME, Mode.MEMORY):
            pprof_sample.value.append(sample.metrics.time)
        else:
            pprof_sample.value.append(sample.metrics.time)
            pprof_sample.value.append(sample.metrics.memory_alloc)
            pprof_sample.value.append(sample.metrics.memory_dealloc)

        # Add locations. Top of the stack first.
        for frame in sample.frames[::-1]:
            pprof_sample.location_id.append(self.get_location(frame))

    def dump(self, stream: BinaryIO) -> None:
        """Dump the pprof protobuf message to the given binary stream."""
        stream.write(self.profile.SerializeToString())



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
import re
from typing import Any, Dict, List, Optional, TextIO, Type

from austin import AustinError

# ---- Custom types ----
ThreadName = str
ProcessId = int
MicroSeconds = int
KiloBytes = int


# ---- Exceptions ----


class InvalidFrame(AustinError):
    """Invalid frame.

    Thrown when attempting to parse a string that is supposed to represent a
    frame, but has the wrong structure.
    """

    pass


class InvalidSample(AustinError):
    """Invalid sample.

    Thrown when attempting to parse a string that is supposed to represent a
    sample, but has the wrong structure.
    """

    pass


# ---- Dataclasses ----


@dataclass(frozen=True)
class Metrics:
    """Austin metrics."""

    time: MicroSeconds = 0
    memory_alloc: KiloBytes = 0
    memory_dealloc: KiloBytes = 0

    def __add__(self, other: "Metrics") -> "Metrics":
        """Add metrics together (algebraically and component-wise)."""
        return Metrics(
            self.time + other.time,
            self.memory_alloc + other.memory_alloc,
            self.memory_dealloc + other.memory_dealloc,
        )

    def copy(self) -> "Metrics":
        """Make a copy of this object."""
        return self + ZERO

    def __str__(self) -> str:
        """Stringify the metrics."""
        return f"{self.time} {self.memory_alloc} {self.memory_dealloc}"


ZERO = Metrics()


@dataclass(frozen=True)
class Frame:
    """Python frame."""

    function: str
    filename: str
    line: int = 0

    @staticmethod
    def parse(frame: str) -> "Frame":
        """Parse the given string as a frame.

        A string representing a frame has the structure

            ``[frame] := <function> (<module>:<line number>)``

        This static method attempts to parse the given string in order to
        identify the parts of the frame and returns an instance of the
        :class:`Frame` dataclass with the corresponding fields filled in.
        """
        if not frame:
            raise InvalidFrame(frame)

        function, _, rest = frame.partition(" (")
        try:
            module, rest = rest.rsplit(":", maxsplit=1)
            line_no = int(rest.rstrip(")"))
        except ValueError:
            raise InvalidFrame(frame)

        return Frame(function, module, line_no)

    def __str__(self) -> str:
        """Stringify the ``Frame`` object."""
        return f"{self.function} ({self.filename}:{self.line})"


@dataclass
class Sample:
    """Austin sample."""

    pid: ProcessId
    thread: ThreadName
    metrics: Metrics
    frames: List[Frame] = field(default_factory=list)

    _ALT_FORMAT_RE = re.compile(r"\);L([0-9]+)")

    @staticmethod
    def parse(sample: str) -> "Sample":
        """Parse the given string as a frame.

        A string representing a sample has the structure

            ``[Process <pid>;]?Thread <tid>[;[frame]]* [metric]*``

        This static method attempts to parse the given string in order to
        identify the parts of the sample and returns an instance of the
        :class:`Sample` dataclass with the corresponding fields filled in.
        """
        if not sample:
            raise InvalidSample(sample)

        if sample[0] != "P":
            raise InvalidSample(f"Sample doesn't begin with process id: {sample}")

        process, _, rest = sample.partition(";")
        try:
            pid = int(process[1:])
        except ValueError:
            raise InvalidSample(f"Sample has invalid process id: {sample}")

        try:
            thread_frames, *metrics = rest.rsplit(maxsplit=3)
            int(metrics[-3])
        except (ValueError, IndexError):
            # Time/memory metrics
            thread_frames, *metrics = rest.rsplit(maxsplit=1)

        if thread_frames[0] != "T":
            raise InvalidSample(f"Sample doesn't have thread id: {sample}")

        thread, _, frames = thread_frames.partition(";")
        thread = thread[1:]

        if frames:
            colon = frames.rfind(";")
            if colon and frames[colon + 1] == "L":
                frames = Sample._ALT_FORMAT_RE.sub(r":\1)", frames)

        try:
            return Sample(
                pid=int(pid),
                thread=thread,
                metrics=Metrics(*(int(metric) for metric in metrics)),
                frames=[Frame.parse(frame) for frame in frames.split(";")]
                if frames
                else [],
            )
        except ValueError as e:
            raise InvalidSample(f"Sample has invalid metric values: {sample}") from e
        except InvalidFrame as e:
            raise InvalidSample(f"Sample has invalid frames: {sample}") from e


@dataclass
class HierarchicalStats:
    """Base dataclass for representing hierarchical statistics.

    The statistics of a frame stack can be thought of as a rooted tree. Hence
    the hierarchy is established by the parent/child relation between the
    nodes in this tree. An instance of this class represents a node, and a
    leaf is given by those instances with an empty ``children`` attribute.

    The ``label`` attribute is used for indexing reasons and therefore should
    be of a hashable type.

    This class overrides the default ``add`` operator so that one can perform
    operations like ``stats1 + stats2``. Note, however, that instances of this
    class are not assumed to be immutable and indeed this operation will modify
    and return ``stats1`` with the outcome of the addition.
    """

    label: Any
    own: Metrics = field(default_factory=Metrics)
    total: Metrics = field(default_factory=Metrics)
    children: Dict[Any, "HierarchicalStats"] = field(default_factory=dict)

    def __lshift__(self, other: "HierarchicalStats") -> "HierarchicalStats":
        """Merge the RHS into the LHS."""
        if self.label != other.label:
            return self

        self.own += other.own
        self.total += other.total

        for frame, child in other.children.items():
            try:
                self.children[frame] << child
            except KeyError:
                self.children[frame] = child

        return self

    def get_child(self, label: Any) -> "HierarchicalStats":
        """Get a child from the children collection."""
        return self.children[label]

    def collapse(self, prefix: str = "") -> List[str]:
        """Collapse the hierarchical statistics."""
        if not self.children:
            return [f";{prefix}{self.label} {self.own}"]

        own = [] if self.own == Metrics() else [f";{self.label} {self.own}"]
        return own + [
            f";{prefix}{self.label}{rest}"
            for _, child in self.children.items()
            for rest in child.collapse()
        ]


@dataclass
class FrameStats(HierarchicalStats):
    """Frame statistics."""

    label: Frame
    height: int = 0
    children: Dict[Frame, "FrameStats"] = field(default_factory=dict)


class ThreadStats(HierarchicalStats):
    """Thread statistics."""

    label: ThreadName
    children: Dict[Frame, FrameStats] = field(default_factory=dict)

    def collapse(self, prefix: str = "T") -> List[str]:
        """Collapse the hierarchical statistics."""
        return super().collapse(prefix)


@dataclass
class ProcessStats:
    """Process statistics."""

    pid: ProcessId
    threads: Dict[ThreadName, ThreadStats] = field(default_factory=dict)

    def collapse(self) -> List[str]:
        """Collapse the hierarchical statistics."""
        return [
            f"P{self.pid}{rest}"
            for _, thread in self.threads.items()
            for rest in thread.collapse()
        ]

    def get_thread(self, thread_name: ThreadName) -> Optional[ThreadStats]:
        """Get thread statistics from this process by name.

        If the given thread name is not registered with this current process
        statistics, then ``None`` is returned.
        """
        return self.threads.get(thread_name)


@dataclass
class AustinStats:
    """Austin statistics.

    This class is used to collect all the statistics about own and total time
    and/or memory generated by a run of Austin. The :func:`update` method is
    used to pass a new :class:`Sample` so that the statistics can be updated
    accordingly.
    """

    child_pid: ProcessId = 0
    processes: Dict[ProcessId, ProcessStats] = field(default_factory=dict)

    def dump(self, stream: TextIO) -> None:
        """Dump the statistics to the given text stream."""
        for _, process in self.processes.items():
            samples = process.collapse()

            if all(sample.endswith(" 0 0") for sample in samples):
                samples = [sample[:-4] for sample in samples]

            for sample in samples:
                stream.write(sample + "\n")

    def get_process(self, pid: ProcessId) -> ProcessStats:
        """Get process statistics for the given PID."""
        return self.processes[pid]

    @classmethod
    def load(cls: Type["AustinStats"], stream: TextIO) -> "AustinStats":
        """Load statistics from the given text stream."""
        stats = cls()
        for line in stream:
            stats.update(Sample.parse(line))
        return stats

    def update(self, sample: Sample) -> None:
        """Update the statistics with a new sample.

        Normally, you would what to generate a new instance of :class:`Sample`
        by using :func:`Sample.parse` on a sample string passed by Austin to
        the sample callback.
        """
        pid = sample.pid or self.child_pid
        thread_stats = ThreadStats(sample.thread, total=sample.metrics)

        # Convert the list of frames into a nested FrameStats instance
        stats: HierarchicalStats = thread_stats
        container = thread_stats.children
        for height, frame in enumerate(sample.frames):
            stats = FrameStats(label=frame, height=height, total=sample.metrics)
            container[frame] = stats
            container = stats.children
        stats.own = stats.total.copy()

        if pid not in self.processes:
            self.processes[pid] = ProcessStats(pid, {sample.thread: thread_stats})
            return

        process = self.processes[pid]
        if sample.thread not in process.threads:
            process.threads[sample.thread] = thread_stats
            return

        process.threads[sample.thread] << thread_stats



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import asdict, dataclass, field
from enum import Enum
import json
from typing import Dict, List, Optional, TextIO

from austin.stats import Frame, InvalidSample, Sample, ZERO

SpeedscopeJson = Dict
SpeedscopeWeight = int
ProfileName = str


class Units(Enum):
    """Metric units."""

    MICROSECONDS = "μs"
    BYTES = "bytes"


@dataclass(frozen=True)
class SpeedscopeFrame:
    """Speedscope Frame object."""

    name: str
    file: str
    line: int


@dataclass
class SpeedscopeProfile:
    """Speedscope Profile object."""

    name: ProfileName
    unit: Units
    startValue: int = 0
    endValue: int = 0
    samples: List[List[int]] = field(default_factory=list)
    weights: List[SpeedscopeWeight] = field(default_factory=list)
    type: str = "sampled"

    def add_sample(self, stack: List[int], weight: SpeedscopeWeight) -> None:
        """Add a sample to the profile."""
        self.samples.append(stack)
        self.weights.append(weight)
        self.endValue += weight


class Speedscope:
    """Speedscope JSON generator."""

    def __init__(self, name: str, indent: Optional[int] = None) -> None:
        self.name = name
        self.indent = indent

        self.profiles: List[SpeedscopeProfile] = []
        self.profile_map: Dict[int, Dict[str, Dict[str, SpeedscopeProfile]]] = {}

        self.frames: List[dict] = []
        self.frame_map: Dict[Frame, int] = {}

    def get_frame(self, frame: Frame) -> int:
        """Get the index of an observed frame."""
        if frame in self.frame_map:
            return self.frame_map[frame]

        index = len(self.frames)
        self.frame_map[frame] = index
        self.frames.append(
            asdict(SpeedscopeFrame(frame.function, frame.filename, frame.line))
        )

        return index

    def get_profile(self, pid: int, thread: str, metric: str) -> SpeedscopeProfile:
        """Get the profile for the given pid, thread and profile metric."""
        prefix = {"t": "Time", "m+": "Memory allocation", "m-": "Memory deallocation"}[
            metric
        ]
        units = Units.BYTES if metric[0] == "m" else Units.MICROSECONDS
        profiles = self.profile_map.setdefault(pid, {}).setdefault(thread, {})
        if metric in profiles:
            return profiles[metric]

        self.profiles.append(
            SpeedscopeProfile(
                name=f"{prefix} profile for {pid}:{thread}", unit=units.value,
            )
        )
        return profiles.setdefault(metric, self.profiles[-1],)

    def add_sample(self, sample: Sample) -> None:
        """Add a sample to the generator."""
        if not sample.frames or sample.metrics == ZERO:
            return

        self.get_profile(sample.pid, sample.thread, "t").add_sample(
            [self.get_frame(frame) for frame in sample.frames], sample.metrics.time
        )

        if sample.metrics.memory_alloc:
            self.get_profile(sample.pid, sample.thread, "m+").add_sample(
                [self.get_frame(frame) for frame in sample.frames],
                sample.metrics.memory_alloc,
            )

        if sample.metrics.memory_dealloc:
            self.get_profile(sample.pid, sample.thread, "m-").add_sample(
                [self.get_frame(frame) for frame in sample.frames],
                -sample.metrics.memory_dealloc,
            )

    def asdict(self) -> SpeedscopeJson:
        """Return the JSON as a Python dictionary."""
        return {
            "$schema": "https://www.speedscope.app/file-format-schema.json",
            "shared": {"frames": self.frames},
            "profiles": sorted(
                [asdict(profile) for profile in self.profiles],
                key=lambda p: p["name"].rsplit(maxsplit=1)[-1],
            ),
            "name": self.name,
            "exporter": "Austin2Speedscope Converter 0.2.0",
        }

    def dump(self, stream: TextIO) -> None:
        """Dump the pprof protobuf message to the given binary stream."""
        json.dump(
            self.asdict(), stream, indent=self.indent,
        )


def main() -> None:
    """austin2speedscope entry point."""
    import os
    from argparse import ArgumentParser

    arg_parser = ArgumentParser(
        prog="austin2speedscope",
        description=(
            "Convert Austin generated profiles to the Speedscope JSON format "
            "accepted by https://speedscope.app. The output will contain a profile "
            "for each thread and metric included in the input file."
        ),
    )

    arg_parser.add_argument(
        "input",
        type=str,
        help="The input file containing Austin samples in normal format.",
    )
    arg_parser.add_argument(
        "output", type=str, help="The name of the output Speedscope JSON file."
    )
    arg_parser.add_argument(
        "--indent", type=int, help="Give a non-null value to prettify the JSON output."
    )

    arg_parser.add_argument("-V", "--version", action="version", version="0.1.0")

    args = arg_parser.parse_args()

    speedscope = Speedscope(os.path.basename(args.input), args.indent)
    try:
        with open(args.input, "r") as fin:
            for line in fin:
                try:
                    speedscope.add_sample(Sample.parse(line))
                except InvalidSample:
                    continue

    except FileNotFoundError:
        print(f"No such input file: {args.input}")
        exit(1)

    with open(args.output, "w") as fout:
        speedscope.dump(fout)


if __name__ == "__main__":
    main()



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import sys
from typing import List

from austin import AustinError, AustinTerminated, BaseAustin
from austin.cli import AustinArgumentParser


class AsyncAustin(BaseAustin):
    """Asynchronous implementation of Austin.

    Implements an ``asyncio`` API for Austin so that it can be used alongside
    other asynchronous tasks.

    The following example shows how to make a simple asynchronous echo
    implementation of Austin that behaves exactly just like Austin.

    Example::

        class EchoAsyncAustin(AsyncAustin):
            def on_ready(self, process, child_process, command_line):
                print(f"Austin PID: {process.pid}")
                print(f"Python PID: {child_process.pid}")
                print(f"Command Line: {command_line}")

            def on_sample_received(self, line):
                print(line)

            def on_terminate(self, data):
                print(data)

        if sys.platform == "win32":
            asyncio.set_event_loop(asyncio.ProactorEventLoop())

        try:
            austin = EchoAsyncAustin()
            asyncio.get_event_loop().run_until_complete(
                austin.start(["-i", "10000", "python3", "myscript.py"])
            )
        except (KeyboardInterrupt, asyncio.CancelledError):
            pass
    """

    async def _read_header(self) -> bool:
        while self._python_version is None:
            line = (await self.proc.stderr.readline()).decode().rstrip()
            if not line:
                return False
            if " austin version: " in line:
                _, _, self._version = line.partition(": ")
            elif " Python version: " in line:
                _, _, self._python_version = line.partition(": ")
        return True

    async def start(self, args: List[str] = None) -> None:
        """Create the start coroutine.

        Use with the ``asyncio`` event loop.
        """
        try:
            self.proc = await asyncio.create_subprocess_exec(
                self.binary_path,
                *(args or sys.argv[1:]),
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
        except FileNotFoundError:
            raise AustinError("Austin executable not found.")

        if not self.proc.stdout:
            raise AustinError("Standard output stream is unexpectedly missing")
        if not self.proc.stderr:
            raise AustinError("Standard error stream is unexpectedly missing")

        self._running = True

        try:
            if not await self._read_header():
                raise AustinError("Austin did not start properly")

            # Austin started correctly
            self._ready_callback(
                *self._get_process_info(
                    AustinArgumentParser().parse_args(args), self.proc.pid
                )
            )

            # Start readline loop
            while self._running:
                data = await self.proc.stdout.readline()
                if not data:
                    break

                self.submit_sample(data)

        finally:
            # Wait for the subprocess to terminate
            self._running = False

            try:
                stderr = (
                    (await asyncio.wait_for(self.proc.stderr.read(), 0.1))
                    .decode()
                    .rstrip()
                )
            except asyncio.TimeoutError:
                stderr = ""
            self._terminate_callback(stderr)

            rcode = await self.proc.wait()
            if rcode:
                if rcode in (-15, 15):
                    raise AustinTerminated(stderr)
                raise AustinError(f"({rcode}) {stderr}")
